from pathlib import Path


from SleePyPhases import SignalPreprocessing as SP, PreManipulation as PM, FeatureExtraction as FE, DataManipulation as DM
from SleePyPhases.phases.Extract import Extract
from SleePyPhases.phases.BuildDataset import BuildDataset
from pyPhases import Phase
from pyPhasesRecordloader import RecordLoader
import os
from pyPhasesML import ModelManager, TrainingSetLoader

class TestRun(Phase):
    useMultiThreading = False
    def main(self):
        import numpy as np
        import pandas as pd
        
        debugConfig = self.getConfig("debugConfig", {})
        print(f"Overwrite Config with debug values in debugConfig: {debugConfig}")

        with self.project:
            self.project.updateConfig(debugConfig)
            loader = self.getConfig("useLoader")
            rl = RecordLoader.get()
            # rl.debug = True

            if not self.useMultiThreading:
                self.log("Disable multi threading for better error messages. To test multi threading, set TestRun.useMultiThreading to True")
                Extract.useMultiThreading = False
                BuildDataset.useMultiThreading = False
            
            self.log("Testing pyPhases settings")
            dataPath = Path(self.getConfig("data-path"))
            if dataPath.exists():
                self.logSuccess(f"Path {dataPath} exists: {dataPath.exists()}")
                
                # Test if directory is writable
                if os.access(dataPath, os.W_OK):
                    self.logSuccess(f"Data Path ({dataPath}) is writable")
                else:
                    self.logError(f"Data Path ({dataPath}) is not writable. Make sure to set the correct path (Config: data-path) and permissions.")
                    exit(1)
            else:
                self.logError(f"Data Path ({dataPath}) does not exist. Please set the correct path (Config: data-path)")
                exit(1)

            testRunData = dataPath / "testrun"
            self.log(f"Creating a new Data Folder ({testRunData.as_posix()}) for the testrun")
            if not testRunData.exists():
                testRunData.mkdir()
            # else:
            #     # remove all files from the testrun folder
            #     for f in testRunData.iterdir():
            #         f.unlink()

            self.log("Creating a new Data Folder for the testrun")

            self.log(f"Test specified RecordLoader: {loader} ({type(rl)})")
            datasetPath = Path(rl.filePath)
            if datasetPath.exists():
                self.logSuccess(f"RL Path ({rl.filePath}) exists: {Path(rl.filePath).exists()}")
            else:
                self.logError(f"RL Path ({rl.filePath}) does not exist. Make sure to download the dataset and set the correct path (Config: {loader}-path)")
                exit(1)

            self.setConfig("testrun", True)

            self.log("Test generating dataset metadata, for first records")
            df = self.getData("metadata", pd.DataFrame)
            self.log(f"Metadata: {df.head()} / {df.shape}")

            if len(df) > 0:
                self.logSuccess("Dataset records loaded")
            else:
                self.logError("Dataset records not loaded")
            # metadata-channels, dataIsFinal

            self.log("Test data version and split")
            dm = self.getData("dataversionmanager")
            self.log(f"Splits (assuming all records exist): {dm.splits}")

            self.log("For debugging purpose, we will reduce the split using the same 2 records for all splits")
            dm.splits["training"][0] = slice(0, 2)
            dm.splits["validation"][0] = slice(0, 2)
            dm.splits["test"][0] = slice(0, 2)

            self.log("Test extracting data with preprocessing defined in (Config: preprocessing)")
            self.log(f"Preprocessing Config: {self.getConfig('preprocessing')}")
            self.log(f"Preprocessing target frequency: {self.getConfig('preprocessing.targetFrequency')}")
            self.log(f"Preprocessing target labelFrequency: {self.getConfig('preprocessing.labelFrequency')}")
            self.log(f"Preprocessing target channels: {self.getConfig('preprocessing.targetChannels')}")
            # check that resample is in each target signal related preprocessing

            self.log(f"Current SignalPreprocessing Class: {type(SP.getInstance(self.getConfig('preprocessing')))}")
            self.log(f"Current PreManipulation Class: {type(PM.getInstance(self.getConfig('preprocessing')))}")
            # self.log(f"Current FeatureExtraction: {type(FE.getInstance(self.getConfig('preprocessing')))}")
            self.log(f"Current SignalPreprocessing: {self.getConfig('preprocessing.stepsPerType')}")
            self.log(f"Current PreManipulation: {self.getConfig('preprocessing.manipulationSteps')}")
            # self.log(f"Current FeatureExtraction: {type(FE.getInstance(self.getConfig('preprocessing')))}")

            memmapOptions = {
                "dtype": self.getConfig("preprocessing.dtype", "float32"),
            }
            dataExporterSignals = self.getData("data-processed", np.memmap, options=memmapOptions)
            dataExporterFeatures = self.getData("data-features", np.memmap, options=memmapOptions)

            self.logSuccess(f"Extract finished: len X: {len(dataExporterSignals)} / len Y: {len(dataExporterFeatures)}")

            self.log("Test data manipulation")
            manipulationSteps = self.getConfig("segmentManipulation")
            self.log(f"Current segmentManipulation: {manipulationSteps}")
            segmentManipulation = DM.getInstance(manipulationSteps, "training", self.project.config)
            self.log(f"Current DataManipulation Class: {type(segmentManipulation)}")
            # segmentManipulation = DM.getInstance(manipulationSteps, "training", self.project.config, recordMetadata=recordMetaData)
            
            X, Y = dataExporterSignals[0], dataExporterFeatures[0]
            X, Y = segmentManipulation((X, Y), None, 0)

            self.logSuccess(f"Segment Manipulation finished: X: {X.shape} / Y: {Y.shape}")

            self.log(f"Test Model: {self.getConfig('modelName')}")

            model = ModelManager.getModel()
            self.log(f"Current Model: {type(model)}")
            self.log(model.summary())
            train = self.project.generateData("dataset-training")
            val = self.project.generateData("dataset-validation")
            trainingsSet = TrainingSetLoader(trainingData=train, validationData=val)
            model.debug = True
            trainedModel = model.train(trainingsSet)

            self.logSuccess("Debug Model trained")




            # allDBRecordIds
            # self.getData("allRecordIds")
            # extract data
            # build data
            # data manipulation
            # load model
            # training
            # validation
            # threshold optimization
            # segment evaluation
            # event evaluation


from SleePyPhases.recordloaders.MedicalDB import MedicalDB
from SleepHarmonizer import RecordLoaderAlice


class RecordLoaderTSM(RecordLoaderAlice):
    config = {}

    def getMetaData(self, recordName):
        metadata = super().getMetaData(recordName)
        del metadata["patientName"]
        del metadata["patientCode"]
        del metadata["patientAdd"]
        del metadata["technician"]

        db = MedicalDB.get()
        row = db.getRecordData(recordName)
        diagnoses = db.getDiagnoses(recordName)
        metadata["psg_type"] = db.getPSGType(recordName)

        if row is not None:
            patient_id, case_id, recordDate, patientAge, gender = row

            metadata["patient"] = patient_id
            metadata["case"] = case_id
            metadata["age"] = patientAge
            metadata["gender"] = gender
            metadata["startRecord"] = recordDate

        if diagnoses is not None:
            metadata.update(diagnoses)
        
        db.close()

        return metadata



from unittest.mock import MagicMock, call, patch, ANY

import numpy as np
import numpy.testing as npt
from pyPhases.test.Mocks import OverwriteConfig
from pyPhases.test.TestCase import TestCase
from pyPhasesRecordloader import Event, RecordSignal, Signal

from SleePyPhases.phases.Extract import Extract, RecordProcessor


class TestExtract(TestCase):
    """ Test Scenario for Extract:
        - Dataset with 10 Recordings [id 0-10]
            - 4 Recordings for training [id 0-3]
            - 3 Recordings for validation [id 4-6]
            - 3 Recordings for test [id 7-9]
    """
    phase = Extract()

    def config(self):
        return {
            "Extract": {
                "useMultiThreading": False,
            },
            "dataversion": {
                "seed": None,
                "split": {
                    "training": ["0:4"],
                    "validation": ["4:7"],
                    "test": ["7:10"],
                },
                "groupBy": None,
                "recordIds": ["0", "1", "2", "3", "4", "5", "6", "7", "8", "9"],
            },
            "preprocessing": {
                "targetFrequency": 1,
                "labelFrequency": 1,
                "dtype": "float16",
                "stepsPerType": {"test": ["normalize"]},
                "targetChannels": ["C1", "C2", "C3"],
                "forceGapBetweenEvents": False,
                "cutFirstAndLastWake": False,
                "fillLastSleepstage": False,
                "featureChannels": [],
                "manipulationSteps": []
            },
            "classification": {
                "name": "arousal",
                "labelNames": "Arousal",
                "classNames": [["None", "Arousal"]],
            },
        }

    C1 = np.array([1, 2, 3, 4, 5])
    C2 = np.array([6, 7, 8, 9, 10])
    C3 = C2

    def getTestSignal(self, recordId=0):
        recordSignal = RecordSignal(recordId=recordId)
        c1 = np.array([1, 2, 3, 4, 5])
        c2 = np.array([6, 7, 8, 9, 10])
        events = [Event("arousal_rera", 0, duration=1), Event("arousal", 1, duration=2), Event("N1", 0, duration=10)]
        recordId = int(recordId)
        recordSignal.addSignal(Signal("C1", c1 + recordId, 1, typeStr="test"))
        recordSignal.addSignal(Signal("C2", c2 + recordId, 1))
        recordSignal.addSignal(Signal("C3", c2 + recordId, 1, typeStr="test"))

        return recordSignal, events

    # def testPrepareConfig(self):
    #     self.assertEqual(self.phase.getConfig("datasetSplits"), ["training", "validation", "test"])

    def testRecordProcessor(self):
        recordSignal, events = self.getTestSignal()

        recordLoader = MagicMock()
        recordLoader.loadRecord.return_value = (recordSignal, events)
        extractor = MagicMock()

        eventManager = MagicMock()
        eventManager.getEventSignalFromList.return_value = {"arousal": np.array([0, 1, 2, 3, 4])}
        stepsPerType = {"test": ["S1", "S2"]}
        preprocessingConfig = {
            "targetFrequency": 1,
            "labelFrequency": 1,
            "dtype": "float16",
            "targetChannels": ["C1", "C2", "C3"],
            "stepsPerType": stepsPerType,
            "cutFirstAndLastWake": False,
            "fillLastSleepstage": False,
            "featureChannels": [],
            "manipulationSteps": [],
        }
        signalLength = 5

        recordProcessor = RecordProcessor(
            recordLoader,
            preprocessingConfig,
            extractor,
            eventManager,
            labelChannels=["SleepArousals"],
            project=self.project,
        )
        signalArray, eventSignal = recordProcessor("id1")

        recordLoader.loadRecord.assert_called_once_with("id1")
        extractor.preprocessingSignal.assert_called_once_with(recordSignal)
        eventManager.getEventSignalFromList.assert_called_once_with(
            events,
            signalLength,
            targetFrequency=1,
            forceGapBetweenEvents=False,
        )

        self.assertEqual(recordSignal.targetFrequency, 1)
        npt.assert_equal(signalArray[:, 0], self.C1)
        npt.assert_equal(signalArray[:, 1], self.C2)
        npt.assert_equal(signalArray[:, 2], self.C3)
        npt.assert_equal(eventSignal, np.array([0, 0, 1, 1, 1]).reshape(-1, 1))

    @patch("pyPhasesRecordloader.RecordLoader.RecordLoader.loadRecord")
    def testExtractTraining(self, loadRecord):
        loadRecord.side_effect = lambda recordId: self.getTestSignal(recordId)

        processedDataSignals = self.getData("data-processed")
        processedDataFeatures = self.getData("data-features")

        self.assertEqual(len(processedDataSignals), 10)
        self.assertEqual(len(processedDataFeatures), 10)

        for i, signalArray in enumerate(processedDataSignals):
            # should have all 3 signale (C1, C2, C3)
            # the recordId is added to the signal
            # normalize is the defaul preprocessing step for C1 and C3
            npt.assert_equal(signalArray[:, 0], np.linspace(0, 1, 5))
            npt.assert_equal(signalArray[:, 1], self.C2 + i)
            npt.assert_equal(signalArray[:, 2], np.linspace(0, 1, 5))

        for eventArray in processedDataFeatures:
            # arousal 0-1, arousal_rera 1-3
            npt.assert_equal(eventArray[:, 0], [1.0, 1.0, 1.0, 0.0, 0.0])

    @patch("SleePyPhases.phases.Extract.RecordProcessor.__call__")
    @patch("pyPhasesML.exporter.MemmapRecordExporter.MemmapRecordExporter.saveAndAppendArray")
    @patch("pyPhasesML.exporter.MemmapRecordExporter.MemmapRecordExporter.finishStream")
    @patch("pyPhasesML.exporter.MemmapRecordExporter.MemmapRecordExporter.exists")
    def testMissingChannels(self, exists, finishStream, saveAndAppendArrayMock, recordProcessorMock, ):
        recordProcessorMock.side_effect = [(1, 2), None, (3, 4), (5, 6), (5, 6), None]
        exists.side_effect = [True]
        self.phase.threads = 3
        removed = self.phase.buildFromRecords(["record1", "record2", "record3", "record4", "record5", "record6"], force=True)

        saveAndAppendArrayMock.assert_has_calls(
            [
                call(ANY, (1, 3)),  # X1
                call(ANY, (2, 4)),  # Y1
                call(ANY, (5, 5)),  # X2
                call(ANY, (6, 6)),  # Y2
            ]
        )
        self.assertEqual(removed, ['record2', 'record6'])

    @patch("SleePyPhases.phases.Extract.RecordProcessor.__call__") # processedsignal, eventsignal
    @patch("pyPhasesML.exporter.MemmapRecordExporter.MemmapRecordExporter.saveAndAppendArray")
    @patch("pyPhasesML.exporter.MemmapRecordExporter.MemmapRecordExporter.finishStream")
    @patch("pyPhasesML.exporter.MemmapRecordExporter.MemmapRecordExporter.exists")
    def testMissingChannelsSingleBatch(self, exists, finishStream, saveAndAppendArrayMock, recordProcessorMock):
        exists.side_effect =  [True]
        recordProcessorMock.side_effect = [(1, 2), None, (3, 4), (5, 6), (5, 6), None]
        self.phase.threads = 3
        removed = self.phase.buildFromRecords(["record1", "record2", "record3", "record4", "record5", "record6"], force=True)
        saveAndAppendArrayMock.assert_has_calls(
            [
                call(ANY, (1, 3)),  # X1
                call(ANY, (2, 4)),  # Y1
                call(ANY, (5, 5)),  # X2
                call(ANY, (6, 6)),  # Y2
            ]
        )
        self.assertEqual(removed, ["record2", "record6"])

    # @OverwriteConfig(datasetSplit="training")
    # @patch("SleePyPhases.phases.Extract.Extract.buildFromRecords")
    # def testDataSetManager(self, buildFromRecords):
    #     buildFromRecords.return_value = [1, 2]
    #     # training: 0, 1, 2, 3
    #     # training after removed: 0, 3

    #     self.phase.registerData("removedRecordIndexes", [1, 2])
    #     dm = self.getData("dataversionmanager")
    #     trainingIds = dm.getRecordsForSplit("training")

    #     self.assertDataEqual("removedRecordIndexes", [1, 2])
    #     self.assertEqual(dm.removedRecords["training"], [1, 2])
    #     self.assertEqual(trainingIds, ["0", "3"])
    #     self.project.unregister("removedRecordIndexes")
    #     self.project.unregister("dataversionmanager")




import pandas as pd
from pyPhases import Phase
from pyPhases.util import BatchProgress
from pyPhasesRecordloader import AnnotationNotFound, RecordLoader
import tqdm


class ExtractEvents(Phase):
    """
    Extract numpy arrays from the raw records in the dataset
    """

    useMultiThreading = True
    threads = None
    validatDataset = True

    def getEventList(self, recordId):
        rl = RecordLoader.get()
        try:
            return rl.getEventList(recordId)
        except AnnotationNotFound:
            return None


    def getEvents(self, records, force=False):
        rl = RecordLoader.get()
        
        bp = BatchProgress(records)
        bp.useMultiThreading = self.useMultiThreading
        if self.threads is not None:
            bp.threads = self.threads

        def combine(events, index):
            combine.events.append([e for e in events if e is not None])
        # def combineFlatten(events, index):
        #     combine.events += [e for ev in events if ev is not None for e in ev]
        # def getEventList(recordId):
        #     rl.getEventList(recordId)
        combine.events = []

        # bp.asyncQueue(rl.getEventList, records, combine)
        bp.start(self.getEventList, records, combine)
        return combine.events
    
    def appendDataframe(self, df, dfToAppend, colMap, offsetStart=None, offsetEnd=None):
        cols = {n: [] for n in colMap}
        if offsetStart is None:
            offsetStart = 0
        if offsetEnd is None:
            offsetEnd = 0
        for e in df.iloc:
            start = e.start + offsetStart
            end = e.end + offsetEnd
            possibleEvents = dfToAppend.query("start < %f and end > %f" % (end, start))
            for name in cols:
                cols[name].append(",".join([ev[colMap[name]] for ev in possibleEvents.iloc]))
        for name in cols:
            df[name] = cols[name]

    def generateData(self, name):
        if name == "events":
            groupedRecords = self.project.getData("allDBRecordIds", list)
            flatten = [r for records in groupedRecords.values() for r in records]
            # extract the data
            events = self.getEvents(flatten)
            self.registerData("events", events)
        if name == "eventDF":
            allRecordEventsBatches = self.getData("events", list)
            events = []
            recordIndex = 0
            for allRecordEvents in tqdm.tqdm(allRecordEventsBatches):
                for recordEvents in allRecordEvents:
                    for event in recordEvents:
                        events.append({
                            "recordIndex": recordIndex,
                            "name": event.name,
                            "start": event.start,
                            "duration": event.duration,
                        })
                    recordIndex += 1
            df = pd.DataFrame.from_dict(events)
            df["end"] = df["start"] + df["duration"]
            dfSleepStage = df.query("name in ['W', 'N1', 'N2', 'N3', 'R']")
            dfArousal = df.query("name in ('arousal', 'arousal_rera')")
            dfLM = df.query("name in ('LegMovement-Left', 'LegMovement-Right')")
            eventResp = df.query("name in ('resp_hypopnea', 'resp_centralapnea', 'resp_obstructiveapnea', 'resp_mixedapnea', 'resp_cheynestokesbreath')")

            self.registerData("eventArousal", dfArousal)
            self.registerData("eventLM", dfLM)
            self.registerData("eventSleep", dfSleepStage)
            self.registerData("eventResp", eventResp)

            self.appendDataframe(dfArousal, dfSleepStage, {"sleepStage": "name"}, offsetStart=-0.5, offsetEnd=0.5)
            self.appendDataframe(dfArousal, dfLM, {"legmovements": "name"}, offsetStart=-0.5, offsetEnd=0.5)
            self.appendDataframe(dfArousal, eventResp, {"resp_event": "name"}, offsetStart=-0.5, offsetEnd=0.5)

            self.registerData("eventDF", dfArousal)
            
            return dfArousal
    
    def appendDataframe(self, df, dfToAppend, colMap, offsetStart=None, offsetEnd=None, idQuery=""):
        cols = {n: [] for n in colMap}
        if offsetStart is None:
            offsetStart = 0
        if offsetEnd is None:
            offsetEnd = 0
        for e in tqdm.tqdm(df.iloc):
            start = e.start + offsetStart
            end = e.end + offsetEnd
            possibleEvents = dfToAppend.query("recordIndex == %i and start < %f and end > %f" % (e.recordIndex, end, start))
            for name in cols:
                cols[name].append(",".join([ev[colMap[name]] for ev in possibleEvents.iloc]))
        for name in cols:
            df[name] = cols[name]
            
    def plotArousalHist(self, df, col, name):
        from matplotlib import pyplot as plt
        
        df[col].plot(kind='hist', bins=100, density=False)
        plt.xlabel(name)
        # plt.ylabel('Density')
        plt.ylabel('Arousal Count')
        plt.title('Distribution of %s (M=%.2f, μ=%.2f, σ=%.2f)'%(name, df["duration"].median(), df["duration"].mean(), df["duration"].std()))



    def main(self):
        from matplotlib import pyplot as plt
        import matplotlib.ticker as ticker
        
        from SleePyPhases.Plot import Plot
        
        # df = self.getData("eventDF")
        df = self.getData("eventDF", list)
        
        modelConfigString = self.project.getDataFromName("eventDF").getTagString()
        logPath = f"eval/{modelConfigString}/"
        self.logPath = logPath
        plot = Plot(logPath)
        plot.createLogFolder()
        plot.plotClassDistribution(df.groupby("name")["recordIndex"].count(), "Arousal/Rera")
        plot.save("arousal-rera")
        plot.plotClassDistribution(df.query("name == 'arousal'").groupby("sleepStage")["recordIndex"].count(), "sleepStage Arousal")
        plot.save("arousal-sleepStage")
        plot.plotClassDistribution(df.query("name == 'arousal_rera'").groupby("sleepStage")["recordIndex"].count(), "sleepStage Rera")
        plot.save("rara-sleepStage")
        plot.plotClassDistribution(df.query("name == 'arousal_rera'").groupby("sleepStage")["recordIndex"].count(), "sleepStage")
        plot.save("sleepStage")
        
        df["type"] = df.apply(lambda row: ("L" if row["legmovements"] != "" else "") + ("R" if row["resp_event"] != "" else ""), axis=1)
        df["first_resp"] = df.apply(lambda row: row["resp_event"].split(",")[0], axis=1)
        
        plot.plotClassDistribution(df.groupby("type")["recordIndex"].count(), "type")
        plot.save("type")
        
        plot.plotClassDistribution(df.query("first_resp != ''").groupby("first_resp")["recordIndex"].count())
        plot.save("first_resp")
        
        
        plot.plotClassDistribution(df.groupby("resp_event")["recordIndex"].count(), "resp_events")
        plot.save("resp_events")
        
        
        # df["duration"].plot(kind='density')
        self.plotArousalHist(df, "duration", "Duration")
        plt.text(10 + 2.5, plt.ylim()[1]*0.9, "10", color='red', ha='center', va='bottom')
        plt.axvline(x=10, color='red')
        plot.save("duration")
        
        self.plotArousalHist(df.query("name == 'arousal_rera'"), "duration", "Rera Duration")
        plt.text(10 + 2.5, plt.ylim()[1]*0.9, "10", color='red', ha='center', va='bottom')
        plt.axvline(x=10, color='red')
        plot.save("rera-duration")
        
        self.plotArousalHist(df.query("name == 'arousal'"), "duration", "Arousal Duration")
        plt.text(10 + 2.5, plt.ylim()[1]*0.9, "10", color='red', ha='center', va='bottom')
        plt.axvline(x=10, color='red')
        plot.save("arousal-duration")
        
        self.plotArousalHist(df.query("name == 'arousal'"), "start", "Arousal Start")
        plot.save("arousal-start")
        
        self.plotArousalHist(df.query("name == 'arousal_rera'"), "start", "Rera Start")
        plot.save("rera-start")
        
        self.plotArousalHist(df, "start", "Start")
        plt.gca().xaxis.set_major_formatter(ticker.FuncFormatter(lambda x, pos: f"{x/3600:.1f}h"))
        plot.save("start")
        
        



import sys

from pyPhases.util.Logger import classLogger
from pypika import Table, PostgreSQLQuery as Query
from pypika.queries import QueryBuilder


@classLogger
class MedicalDB:
    connection = None
    client = None
    assocCursor = None
    _instance = None

    config = {
        "user": "root",
        "password": "example",
        "host": "localhost",
        "port": 5432,
        "database": "data",
    }

    @classmethod
    def get(cls) -> "MedicalDB":
        if cls._instance is None:
            cls._instance = cls.__new__(cls)
            cls._instance.connect()
        return cls._instance

    def connect(self):

        import psycopg2

        try:
            c = MedicalDB.config
            conn = psycopg2.connect(
                "host=%s port=%i dbname=%s user=%s password=%s"
                % (
                    c["host"],
                    c["port"],
                    c["database"],
                    c["user"],
                    c["password"],
                )
            )

        except psycopg2.Error as e:
            self.logError(f"Error connecting to Postgres Platform: {e}")
            sys.exit(1)
        self.conn = conn
        self.cursor = conn.cursor()

    def execute(self, sql, **vars):
        if isinstance(sql, QueryBuilder):
            sql = sql.get_sql()
        return self.cursor.execute(sql, **vars)

    def executeAndFetch(self, sql, **vars):
        self.execute(sql, **vars)
        return self.cursor.fetchall()

    def commit(self):
        self.conn.commit()
    
    def executeAndFetchOne(self, sql, **vars):
        self.execute(sql, **vars)
        return self.cursor.fetchone()

    def close(self):
        if self.connection is not None:
            self.connection.close()

    def fillRecordId(self, id):
        return id

    def getRecordData(self, recordId):
        record = Table("record", schema=self.config["schema"])
        patient = Table("patient", schema=self.config["schema"])
        medCase = Table("medical_case", schema=self.config["schema"])
        q = (
            Query.from_(record)
            .select(patient.id_patient, record.id_medical_case, record.start, patient.date_of_birth, patient.gender)
            .join(medCase).on(record.id_medical_case == medCase.id_medical_case)
            .join(patient).on(patient.id_patient == medCase.id_patient)
            .where(record.id_record == recordId)
        )
        row = self.executeAndFetchOne(q)

        if row is None:
            return None
        
        row = list(row)
        age = None if row[2] or row[3] is None else row[2] - row[3]
        row[3] = age

        return row

    def getDiagnoses(self, recordId):
        diag = Table("diagnosis", schema=self.config["schema"])
        record = Table("record", schema=self.config["schema"])
        medCase = Table("medical_case", schema=self.config["schema"])
        q = Query.from_(diag)\
            .select("diagnosis_category_id", "diagnosis")\
            .join(medCase).on(diag.id_medical_case == medCase.id_medical_case)\
            .join(record).on(medCase.id_medical_case == record.id_medical_case)\
            .where(record.id_record == recordId)\
            .get_sql()
        
        diagnoses = self.executeAndFetch(q)

        diagnoseCats = {d[0]: d[1] for d in diagnoses}
        return diagnoseCats

    def getPSGType(self, recordId):

        record = Table("record", schema=self.config["schema"])
        q = (
            Query.from_(record)
            .select("psg_config")
            .where(record.id_record == recordId)
        )
        r = self.executeAndFetchOne(q)

        return r[0] if bool(r) else ""




from itertools import chain
from pyPhases import Phase
from tqdm import tqdm

from pyPhasesRecordloader import AnnotationNotFound, ChannelsNotPresent, ParseError

class ExportMetadataToMeticalDB(Phase):
    """export extracted metadata from psg files to the patient db"""

    exportTo = "db"

    def main(self):
        import pandas as pd
        from pypika import PostgreSQLQuery, Schema
        from SleepHarmonizer.recordloaders.AliceTextReportLoader import AliceTextReportLoader

        from SleePyPhases.recordloaders.MedicalDB import MedicalDB

        metadata = self.getData("metadata", pd.DataFrame)
        relevantData = [v for v in AliceTextReportLoader.relevantRows.values() if v != ""]
        # recordMetadataRows = []
        # for dataRow in metadata.iloc:

        #     idDict = {
        #         "acquisition_id": dataRow["recordId"],
        #         "patient_psn": dataRow["patient"] if "patient" in dataRow else None,
        #         "bmi": dataRow["bmi"] if "bmi" in dataRow else None,
        #         "domino": self.getConfig("useLoader") == "tsm-domino"
        #     }
        #     for col in cols:
        #         if col in dataRow:
        #             idDict[col] = dataRow[col]
        #     recordMetadataRows.append(idDict)

        
        # channelMetadata = self.getData("metadata-channels", pd.DataFrame)
        # features = self.getData("features", pd.DataFrame)

        # for dataRow in list(channelMetadata.iloc) + list(features.iloc):
        #     pass

        # if self.exportTo == "csv":
        #     df = pd.DataFrame.from_records(recordMetadataRows, index="acquisition_id")
        #     df.to_csv("recordMetadata.csv")
        # else:
        MedicalDB.config = self.getConfig("medicalDB")
        medDB = MedicalDB.get()
        medDB.connect()
        schema = Schema(self.getConfig("medicalDB.schema"))

        patientTable = schema.patient
        caseTable = schema.medical_case
        recordTable = schema.record
        recordMetaTable = schema.record_somnometric
        channelTable = schema.channel
        channelMetaTable = schema.channel_somnometric
        somnoemetricTable = schema.somnometric_type
        
        for rowIndex, row in tqdm(metadata.iterrows(), total=len(metadata)):
            # df = metaData.items()
            recordId = row["recordId"]
            channelData = self.getData("metadata-channels", pd.DataFrame, recordId=recordId)
            features = {}
            try:
                features = self.getData("features", pd.DataFrame, recordId=recordId)
            except ChannelsNotPresent:
                pass
            except AnnotationNotFound:
                pass
            except ParseError:
                pass

            missingCaseCounts = {}
            def getMissingCid(pid):
                if pid in missingCaseCounts:
                    missingCaseCounts[pid] += 1
                else:
                    missingCaseCounts[pid] = 0
                return "missing_%s_%s" % (pid, missingCaseCounts[pid]) 
            
            patientId = row["patient"] if "patient" in row and not pd.isna(row["patient"]) else "missing_%s" % rowIndex
            caseId = row["case"] if "case" in row and not pd.isna(row["case"]) else getMissingCid(patientId)
            recordId = row["recordId"]
            
            recordEntries = {
                "recordId": "id_record", 
                "start": "start", 
                "dataCount": "data_count",
                "psg_setup": "psg_config",
                "lightOff": "light_off",
                "lightOn": "light_on",
            }
            removeEntries = ["annotationExist", "channelMissing", "samplingRateCheck"]
            removeEntries += ["diagnoses", "normalbefund", "diagnosen_rdi_lt5_lt15", "sbas_nachrdi_lt5_lt15", "andere_hinweise", "sonstige", "rbd", "hypersomnie", "insomnie", "rls_plmd"]
            removeEntries += ["startRecord", "comment"]
            removeEntries += ["age", "gender"]
            removeEntries += ["case", "patient", "psg_type"]

            dateEntries = ["start"]

            for d in dateEntries:
                if pd.notna(row.loc[d]):
                    row.at[d] = row.loc[d].isoformat()      
                          
            for r in removeEntries:
                if r in row:
                    del row[r]

            record = {target: row[src] for src, target in recordEntries.items() if src in row and not pd.isna(row[src])}

            for src,target in recordEntries.items():
                if src in row:
                    del row[src]

            # insert patient information
            q = PostgreSQLQuery.into(patientTable)\
                .columns(patientTable.id_patient)\
                .insert(patientId)\
                .on_conflict().do_nothing()
            medDB.execute(str(q))

            # insert case information
            q = PostgreSQLQuery.into(caseTable)\
                .columns("id_medical_case", "id_patient")\
                .insert(caseId, patientId)\
                .on_conflict().do_nothing()
            medDB.execute(str(q))

            # insert record information
            record["id_medical_case"] = caseId 
            q = PostgreSQLQuery.into(recordTable)\
                .columns(*list(record.keys()))\
                .insert(*list(record.values()))\
                .on_conflict(recordTable.id_record)
            
            for col, val in record.items():
                if col != "id_record" and not pd.isna(val):
                    q = q.do_update(col, val)
            medDB.execute(str(q))

            def getTypeAndValue(val):
                if isinstance(val, bool):
                    return "float", 1 if val else 0
                
                dtype = "float" if val.dtype == float else "string"

                return dtype, val
            
            def addMetricType(metricName, mType):
                
                mType, val = getTypeAndValue(mType)

                q = PostgreSQLQuery.into(somnoemetricTable)\
                    .columns("id_somnometric_type", "valuetype")\
                    .insert(metricName, mType)\
                    .on_conflict("id_somnometric_type")\
                    .do_update("valuetype", mType)
                    # .returning('id_somnometric_type')
                return medDB.execute(str(q))

            # delete all record metadata
            q = PostgreSQLQuery.from_(recordMetaTable)\
                .delete()\
                .where(recordMetaTable.id_record == recordId)
                # where source == psg
            medDB.execute(str(q))

            
            # add all the metadata types (somnometrics types)
            # channelMetrics = [col for col in channelData.columns if col not in ["label", "signalName", "recordId"]]
            # for metricName in channelMetrics:
                # addMetricType(metricName, channelData[metricName].dtype)

            if len(features) > 0:
                featureMetrics = [col for col in features.columns if col not in ["channel", "recordId"]]
                for metricName in featureMetrics:
                    addMetricType(metricName, features[metricName])

            for channel in channelData.iloc:
                if "signalName" not in channel:
                    continue
                chanelDict = {
                    "name": channel["signalName"],
                    "frequency": channel["sample_rate"], # sample_frequency
                    "prefilter": channel["prefilter"],
                    "physical_max": channel["physical_max"],
                    "physical_min": channel["physical_min"],
                    "digital_max": channel["digital_max"],
                    "digital_min": channel["digital_min"],
                    "prefilter": channel["prefilter"],
                    "transducer": channel["transducer"],
                    "type": channel["type"],
                    "id_record": recordId
                }
                
                q = PostgreSQLQuery.into(channelTable)\
                    .columns(*list(chanelDict.keys()))\
                    .insert(*list(chanelDict.values()))
                channelId = medDB.execute(str(q))


                # add channel specific features
                if len(features) == 0:
                    continue

                channelFeatures = features[features["channel"] == channel["signalName"]]

                if len(channelFeatures) > 0:
                    # each channel/recording should only have a single row
                    channelFeatures = channelFeatures.iloc[0]
                    for feat in featureMetrics:
                        val = channelFeatures[feat]
                        if not pd.isna(val):
                            typestr, val = getTypeAndValue(val)
                            
                            q = PostgreSQLQuery.into(channelMetaTable)\
                                .columns(["id_channel", "id_somnometric_type", f"value_{typestr}"])\
                                .insert([channelId, feat, val])
                            medDB.execute(str(q))


                # q = PostgreSQLQuery.from_(recordMetaTable)\
                #     .delete()\
                #     .where(recordMetaTable.id_somnometric_type == metricName)\
                #     .where(recordMetaTable.id_record == recordId)
                # medDB.execute(str(q))

                # q = PostgreSQLQuery.into(recordMetaTable)\
                #     .columns("id_somnometric_type", "id_record", f"value_{mType}")\
                #     .insert(metricName, recordId, value)
                # medDB.execute(str(q))
            

            q = PostgreSQLQuery.from_(channelTable)\
                .delete()\
                .where(channelTable.id_record == recordId)
            medDB.execute(str(q))

            if "channels" not in row:
                continue

            for channel in row["channels"]:
                channelName = channel["signalName"]

                chanelDict = {
                    "name": channelName,
                    "frequency": channel["sample_rate"], # sample_frequency
                    "prefilter": channel["prefilter"],
                    "physical_max": channel["physical_max"],
                    "physical_min": channel["physical_min"],
                    "digital_max": channel["digital_max"],
                    "digital_min": channel["digital_min"],
                    "prefilter": channel["prefilter"],
                    "transducer": channel["transducer"],
                    "type": channel["type"],
                    "id_record": recordId
                }
                q = PostgreSQLQuery.into(channelTable)\
                    .columns(*list(chanelDict.keys()))\
                    .insert(*list(chanelDict.values()))
                medDB.execute(str(q))

            if rowIndex % 25 == 0:
                medDB.commit()
            # table = schema.acq_metadata

            # checkQuery = PostgreSQLQuery.from_(table).select(1).where(table.acquisition_id == metaData['acquisition_id'])
            # medDB.execute(str(checkQuery))
            # exist = medDB.executeAndFetchOne(checkQuery)
            # if exist:
            #     q = PostgreSQLQuery.update(table).where(table.acquisition_id == metaData['acquisition_id'])
            #     for f, v in metaData.items():
            #         q = q.set(f, v)
            # else:
            #     q = PostgreSQLQuery.into(table).columns(existingColumns).insert(values)
            # medDB.execute(q)
        medDB.close()



import numpy as np
from pyPhases import Phase
from pyPhasesML import Model, ModelManager
from tqdm import tqdm

from SleePyPhases.DataManipulation import DataManipulation
from SleePyPhases.EventScorer import EventScorer
from SleePyPhases.MultiScorer import MultiScorer


class ThresholdOptimisation(Phase):
    allowTraining = False

    def getModel(self) -> Model:
        modelState = self.project.getData("modelState", Model, generate=self.allowTraining)
        model = ModelManager.getModel(True)
        model.build()
        model.loadState(modelState)
        return model

    def calculateValidationData(self):
        model = self.getModel()
        useTrainValForOptimisation = self.getConfig("optimizeOn")
        assert useTrainValForOptimisation in ["validation", "training", "trainval"]
        validationData = self.project.generateData(f"dataset-{useTrainValForOptimisation}")

        daSteps = self.getConfig("eventEval.augmentationCountThatShouldNotBeOptimized", 0)
        manipulationAfterPredict = self.getConfig("eventEval.manipulationAfterPredict", False)
        manipulationAfterPredict = manipulationAfterPredict[0:daSteps]
        da = DataManipulation.getInstance(manipulationAfterPredict, "test", self.project.config)

        prediction = []
        truth = []
        for i in tqdm(range(0, len(validationData))):
            x, t = validationData[i]
            p = model.predict(x, get_likelihood=True, returnNumpy=False)
            p, t = da((p, t))
            prediction.append(p)
            truth.append(t)

        # we assume output is reduces to bachsize 1 during augmentation (1, X, channels)
        prediction = np.concatenate(prediction, axis=1)
        truth = np.concatenate(truth, axis=1)

        return prediction, truth

    def calculateThreshold(self):
        import scipy
        
        # optimize threshold for given metric
        thresholdMetrics = self.getConfig("thresholdMetric", None)
        thresholdMetrics = thresholdMetrics if isinstance(thresholdMetrics, list) else [thresholdMetrics] 

        classNums = self.getConfig("model.multiclassificationSize", [2])
        labelNames = self.getConfig("classification.labelNames")
        ignoreIndex = self.getConfig("classification.ignoreIndex", -1)

        scorer = MultiScorer(classNums, thresholdMetrics, scorerNames=labelNames, ignoreClasses=[ignoreIndex], scorerClass=EventScorer)
        for s in scorer.scorer.values():
            s.majorityVote = self.getConfig("eventEval.tpStrat", "overlap") == "majority"
            s.noTN = self.getConfig("eventEval.tnStrat", "eventcount") == "noTN"
            s.trace = True


        manipulationAfterPredict = self.getConfig("eventEval.manipulationAfterPredict", False)
        daSteps = self.getConfig("eventEval.augmentationCountThatShouldNotBeOptimized", 0)
        da = DataManipulation.getInstance(manipulationAfterPredict, "test", self.project.config, threshold=0.5)

        # get Validation data
        prediction, truth = self.getData("validationResult", tuple)

        def score(threshold, p, t, scorerIndex, scorerName, optimizeMetric):
            biggerIsBetter = scorer.getMetricDefinition(optimizeMetric)[2]
            if threshold >= 1 or threshold <= 0:
                return np.inf

            negate = -1 if biggerIsBetter else 1
            da.threshold = threshold[0]
            config = manipulationAfterPredict[scorerIndex] if isinstance(manipulationAfterPredict[0], list) else manipulationAfterPredict
            config = config[daSteps:]
            p, t = da((p.copy(), t.copy()), config)
            scorer.scorer[scorerName].threshold = threshold[0]
            r = scorer.scoreSingle(scorerIndex, t, p)

            self.log(f"Optimizing threshold by {optimizeMetric}: t = {threshold} => {r[optimizeMetric]} ")

            return r[optimizeMetric] * negate
        
        thresholds = []
        for scorerIndex, (scorerName, thresholdMetric) in enumerate(zip(labelNames, thresholdMetrics)):
            self.log(f"Optimize threshold for {scorerName} {thresholdMetric}")
            t = scipy.optimize.fmin(score, args=(prediction, truth, scorerIndex, scorerName, thresholdMetric), x0=0.3, xtol=0.01, ftol=0.01)
            self.logSuccess(f"Optimized threshold for {scorerName} {thresholdMetric}: {t}")
            thresholds.append(t)

        return thresholds

    def generateData(self, name):
        if name == "validationResult":
            valData = self.calculateValidationData()
            self.registerData("validationResult", valData)
        if name == "threshold":
            thresholdMetric = self.getConfig("thresholdMetric", None)

            if thresholdMetric is not None:
                threshold = self.calculateThreshold()
            else:
                threshold = 0.5
            self.registerData("threshold", threshold)

    def main(self):
        self.getData("threshold", list)



import unittest
import numpy as np

from SleePyPhases.EventScorer import EventScorer

class TestEventScorer(unittest.TestCase):

    def test_scoreEventsConfusion(self):
        # Define your test inputs and expected output here
        truth = np.array([0, 1, 1, 0, 0, 1, 1, 1, 1, 0, 0, 0])
        prediction = np.array([0, 1, 1, 1, 0, 0, 0, 1, 0, 0, 0, 0])

        numClasses = 2
        classNames = ['class1', 'class2']
        eventScorer = EventScorer(numClasses=numClasses, classNames=classNames)

        # Set other parameters if needed, like threshold and majorityVote

        # Call the method to be tested
        confusion_matrix = eventScorer.scoreEventsConfusion(truth, prediction)

        # Define the expected confusion matrix based on your inputs
        expected_confusion = np.array([
            [3, 0],
            [0, 2],
        ])

        # Compare the result with the expected value
        np.testing.assert_array_equal(confusion_matrix, expected_confusion)
    
    def test_scoreEventsConfusionFN(self):
        # Define your test inputs and expected output here
        truth =      np.array([0, 1, 1, 0, 0, 1, 1, 1, 1, 0, 1, 1])
        prediction = np.array([0, 1, 1, 1, 0, 0, 0, 1, 0, 0, 0, 0])

        numClasses = 2
        classNames = ['class1', 'class2']
        eventScorer = EventScorer(numClasses=numClasses, classNames=classNames)

        # Set other parameters if needed, like threshold and majorityVote

        # Call the method to be tested
        confusion_matrix = eventScorer.scoreEventsConfusion(truth, prediction)

        # Define the expected confusion matrix based on your inputs
        expected_confusion = np.array([
            [3, 0],
            [1, 2],
        ])

        # Compare the result with the expected value
        np.testing.assert_array_equal(confusion_matrix, expected_confusion)

    def test_scoreEventsConfusion_majority(self):
        # Define your test inputs and expected output here
        truth =      np.array([0, 1, 1, 0, 0, 1, 1, 1, 1, 0, 0, 0])
        prediction = np.array([0, 1, 1, 1, 0, 0, 0, 1, 0, 0, 0, 0])

        numClasses = 2
        classNames = ['class1', 'class2']
        eventScorer = EventScorer(numClasses=numClasses, classNames=classNames)
        eventScorer.majorityVote = True

        # Set other parameters if needed, like threshold and majorityVote

        # Call the method to be tested
        confusion_matrix = eventScorer.scoreEventsConfusion(truth, prediction)

        # Define the expected confusion matrix based on your inputs
        expected_confusion = np.array([
            [3, 0],
            [1, 1],
        ])

        # Compare the result with the expected value
        np.testing.assert_array_equal(confusion_matrix, expected_confusion)
    
    # def test_scoreEventsConfusionFN_majority(self):
    #     # Define your test inputs and expected output here
    #     truth = np.array([0, 1, 1, 0, 0, 1, 1, 1, 1, 0, 1, 1])
    #     prediction = np.array([0, 1, 1, 1, 0, 0, 0, 1, 0, 0, 0, 0])  # Values beyond 4 will be ignored

    #     numClasses = 2
    #     classNames = ['class1', 'class2']
    #     eventScorer = EventScorer(numClasses=numClasses, classNames=classNames)
    #     eventScorer.majorityVote = True

    #     # Set other parameters if needed, like threshold and majorityVote

    #     # Call the method to be tested
    #     confusion_matrix = eventScorer.scoreEventsConfusion(truth, prediction)

    #     # Define the expected confusion matrix based on your inputs
    #     expected_confusion = np.array([
    #         [3, 0],
    #         [1, 2],
    #     ])

    #     # Compare the result with the expected value
    #     np.testing.assert_array_equal(confusion_matrix, expected_confusion)

    # # def test_scoreEventsConfusion(self):
    # #     truth = np.array([0, 0, 1, 1, 0, 1, 0, 1, 0])
    # #     prediction = np.array([0, 1, 1, 0, 0, 1, 1, 0, 0])

    # #     (tn, fn), (fp, tp) = EventScorer(numClasses=2).scoreEventsConfusion(truth, prediction)
    # #     self.assertEqual(tn, 4)
    # #     self.assertEqual(fn, 1)
    # #     self.assertEqual(fp, 0)
    # #     self.assertEqual(tp, 2)

    # # def test_scoreMetrics(self):
    # #     truth = np.array([0, 0, 1, 1, 0, 1, 0, 1, 0])
    # #     prediction = np.array([0, 1, 1, 0, 0, 1, 1, 0, 0])

    # #     metrics = EventScorer(numClasses=2).scoreMetrics(truth, prediction)
    # #     self.assertEqual(metrics["kappa"], 0.696)
    # #     self.assertEqual(metrics["accuracy"], 0.857)



import shutil
from pathlib import Path
from typing import Iterable

import numpy as np

import yaml
from pyPhases import Phase
from pyPhasesRecordloader import RecordSignal, Signal

from SleePyPhases import Plot
from SleePyPhases.DataManipulation import DataManipulation
from SleePyPhases.Reporter import Reporter


class EvalReport(Phase):
    allowTraining: bool = True
    exampleSkipClass = ["0-0"]
    examplePaddingInS = 10
    highlight = None
    exampleSecondsPerInch = 2

    def getExternalTestset(self, split="test"):
        with self.project:
            evalConfig = self.getConfig("evalOn")

            # try not to update dataversion, rather replace it completly
            if "dataversion" in evalConfig:
                self.project.config["dataversion"] = {}

            self.project.config.update(evalConfig)
            self.project.trigger("configChanged", None)
            dataset = self.project.getData(f"dataset-{split}", list)
            dm = self.getData("dataversionmanager")
            recordsMap = dm.getRecordsForSplit(split)

        return dataset, dm, recordsMap

    def createLogFolder(self, path):
        Path(path).mkdir(parents=True, exist_ok=True)
        Path(path).joinpath("/examples").mkdir(parents=True, exist_ok=True)

    def addAsset(self, name, extension="png", bbox_inches="tight", dpi=300):
        import matplotlib.pyplot as plt
        fullName = name + "." + extension
        # assetPath = Reporter.assetPath + "/" + fullName
        fullPath = self.evalPath + fullName
        self.createLogFolder(Path(fullPath).parent)
        plt.savefig(fullPath, bbox_inches=bbox_inches, dpi=dpi)
        plt.close()

    def plotRecordSignal(
        self, label, classExamples, recordSignal: RecordSignal, fileName, sliceValues=None, highlights=None, offset=0, classNames=None, threshold=None, labelName=""
    ):

        classificationNames = {}
        for i, name1 in enumerate(classNames):
            for j, name2 in enumerate(classNames):
                classificationNames[f"{i}-{j}"] = f"{name1} as {name2}"

        # classificationName = self.getConfig("classification.name")
        labelFrequency = self.getConfig("preprocessing.labelFrequency")
        samplingrate = recordSignal.targetFrequency

        for classIndex, examples in classExamples.items():
            for eIndex, sliceValues in enumerate(examples):
                className = classificationNames["-".join(classIndex.split("-")[1:])]
                if classIndex in self.exampleSkipClass:
                    continue
                paddingInS = self.examplePaddingInS
                padding = paddingInS * samplingrate
                highlights = []

                sliceValues = sliceValues[0] + offset, sliceValues[1] + offset

                sliceValues = sliceValues[0], sliceValues[0] + 50 * 100

                if self.highlight is not None:
                    s, e = self.highlight
                    s += paddingInS
                    e = e + paddingInS if e > 0 else e - paddingInS
                    highlights.append((s, e))

                plot = Plot(self.evalPath)
                from matplotlib import pyplot as plt

                def mergePrediction(signal, axs, index):
                    if signal.name in labelNames:
                        axs[index+1] = axs[index]
                
                plot.on("beforePlotSignal", mergePrediction)

                labelNames = self.getConfig("classification.labelNames")
                
                def addThreshold(signal, axs, index):
                    if signal.name in labelNames:
                        axs[index].axhline(y=threshold[0], color="red")
                        axs[index].text(1, threshold[0] + 0.02, "Threshold", color="red")
                
                plot.on("afterPlotSignal", addThreshold)

                plot.plotRecordSignal(
                    recordSignal,
                    sliceValues,
                    resample=False,
                    padding=padding,
                    secondsPerInch=self.exampleSecondsPerInch,
                    highlights=highlights,
                    title="Example for %s: %s in record '{record}' from {from}" % (label, className),
                )
                if threshold is not None:
                    fig, ax = plt.subplots()
                    from matplotlib import pyplot as plt
                    plt.axhline(y=threshold[0], color="red")
                    plt.text(1, threshold[0] + 0.02, "Threshold", color="red")

                plot.plot.save(f"examples/{eIndex}-{fileName}-{label}-{className}")

    def plotExamples(self, recordStats, segmentStats, name="", labelName=""):
        channelNames = self.getConfig("evalChannelNames")

        exampleCount = 3
        # pick random records
        randomRecordIndexes = np.random.choice(len(recordStats), exampleCount, replace=False)

        testData, dm, recordsMap = self.getExternalTestset()
        recordIds = dm.getRecordsForSplit("test")

        self.setConfig("BuildDataset.useMultiThreading", False)
        da = DataManipulation.getInstance([], "test", self.project.config)

        for i, recordIndex in enumerate(randomRecordIndexes):
            # create recordsignal for processed data
            samplingrate = self.getConfig("preprocessing.targetFrequency")
            X, Y = testData[recordIndex]
            classExamples = recordStats[recordIndex][f"examples_{labelName}"]

            if not isinstance(X, np.ndarray):
                X = X.numpy()

            if not isinstance(Y, np.ndarray):
                Y = Y.numpy()
                

            recordSignal = RecordSignal.fromArray(
                X, sourceFrequency=samplingrate, targetFrequency=samplingrate, names=channelNames, transpose=True
            )

            # add labels

            recordSignal.recordId = recordIds[recordIndex]
            
            labelNames = self.getConfig("classification.labelNames")
            mainLabelName = labelName
            mainLabelIndex = labelNames.index(labelName)

            labelOrdered = {i:label for i, label in enumerate(labelNames) if label != mainLabelName}
            labelOrdered[mainLabelIndex] = mainLabelName


            
            for labelIndex, label in labelOrdered.items():
                labelSignal = Signal(label, Y[:, labelIndex], samplingrate)
                # labelSignal = Signal(classificationName, Y[:, 0], samplingrate)
                recordSignal.addSignal(labelSignal)


                prediction = segmentStats[label][recordIndex]["prediction"]
                # add prediction
                offset = (len(Y[:, 0]) - prediction.shape[0]) // 2
                prediction, _ = da._fixeSizeSingleChannel(prediction.reshape(1, -1, 1), 2**21, fillValue=0, position="center")
                predictionSignal = Signal(f"{label} Prediction", prediction.reshape(-1), samplingrate)
                recordSignal.addSignal(predictionSignal)

            classNames = [f"No {labelName}", labelName]
            classExamples = {index: [(c[0]+offset, c[1]+offset) for c in ce] for index, ce in classExamples.items()}
            self.plotRecordSignal(labelName, classExamples, recordSignal, f"example-{i}-processed", classNames=classNames, offset=offset, threshold=self.threshold[mainLabelIndex], labelName=labelName)
        
    def main(self):
        import pandas as pd
        from SleePyPhases.Plot import Plot
        
        threshold = self.getConfig("fixedThreshold", False)
        self.threshold = threshold or self.getData("threshold", float)

        modelConfigString = self.project.getDataFromName("eventResults").getTagString()
        evalPath = f"eval/{modelConfigString}/"

        Reporter.basePath  = evalPath
        self.evalPath = evalPath
        self.reporter = Reporter("EvalReport")
        self.reporter.createFolder()
        # self.reporter.save()

        self.copyRawData()
        # self.modelInfo()

        # self.log("save report to: %s" % self.reporter.getFilePath())

        segmentResultsRecords, segmentResults = self.getData("evalResults", list)
        df = pd.DataFrame(segmentResultsRecords).transpose()

        # drop generated columns if they exist
        drop_colums = ["truth", "prediction", "all_values", "pos_values", "ign_values", "neg_values"]
        for col in drop_colums:
            if col in df.columns:
                df = df.drop(columns=[col])

        df.to_csv(f"{evalPath}recordResultsSegment.csv", index=False)

        self.logSuccess(f"result segments: {segmentResults}")
        resultRows, result = self.getData("eventResults", list)
        self.logSuccess(f"result events: {result}")
        df = pd.DataFrame(resultRows)

        
        metrics = self.getConfig("trainingParameter.validationMetrics")
        evalMetrics = self.getConfig("eval.metrics")
        labelNames = self.getConfig("classification.labelNames")

        results = {}
        allMetrics = {mA for m in metrics for mA in m} | {mA for m in evalMetrics for mA in m}
        for labelName in labelNames:
            for m in allMetrics:
                metricName = f"{m}_{labelName}"
                if metricName in result:
                    results[f"{metricName}-ev"] = result[metricName]
                if metricName in segmentResults:
                    results[f"{metricName}-seg"] = segmentResults[metricName]

        relMetaData = self.getConfig("eval.clinicalMetrics")
        for m in relMetaData:
            df[f"{m}-prediction-diff"] = df[f"{m}-prediction"] - df[f"{m}-truth"]
            df[f"{m}-prediction-error"] = df[f"{m}-prediction-diff"].abs()

        df.to_csv(f"{evalPath}recordResultsEvents.csv", index=False)

        self.logSuccess(f"Results: {results}")
        results = {k: float(v) for k, v in results.items() if not isinstance(v, Iterable)}
        with open(f"{evalPath}results.yml", "w") as file:
            yaml.dump(results, file)

        self.project.saveConfig(f"{evalPath}project.config")
        self.log(f"Eval finished: {evalPath}")

    def copyRawData(self):
        modelConfigString = self.project.getDataFromName("modelState").getTagString()
        trainingLogPath = f"logs/{modelConfigString}/"
        shutil.copyfile(trainingLogPath + "log.csv", Reporter.basePath + "training.log")
        shutil.copyfile(trainingLogPath + "project.config", Reporter.basePath + "project.json")

    def trainingPlot(self, logFile):
        import matplotlib.pyplot as plt
        import pandas as pd

        self.metrics = self.getConfig("trainingParameter.validationMetrics") 
        df = pd.read_csv(logFile)
        plt.plot(df["epoch"], df["loss"], label="Training Loss")
        plt.plot(df["epoch"], df[self.metrics[0]], label=self.metrics[0])
        self.reporter.addAsset("trainingsProcess")

    def modelInfo(self):
        model = self.project.getPhase("Eval").getModel()
        sum = model.summary()
        self.reporter.addVar("modelName", self.getConfig("modelName"))
        self.reporter.addVar("parameter", model.parameter, "humannumber")
        self.reporter.addVar("modelSummary", sum)



from pathlib import Path

from pyPhases import Phase
from pyPhasesML import DataversionManager, Scorer
from pyPhasesRecordloader import RecordLoader

from SleePyPhases.recordloaders.MedicalDB import MedicalDB


class Setup(Phase):
    def prepareConfig(self):
        numLabels = len(self.getConfig("classification.labelNames"))
        self.setConfig("numLabels", numLabels)

        numClasses = [len(c) for c in self.getConfig("classification.classNames")]
        self.setConfig("numClasses", numClasses)
        
        defaultValidationbatchManipulation = self.getConfig("segmentManipulationEval", self.getConfig("segmentManipulation"))
        self.setConfig("segmentManipulationEval", defaultValidationbatchManipulation)
        
        # alice loader
        RecordLoader.registerRecordLoader("RecordLoaderTest", "SleePyPhases.recordloaders")
        RecordLoader.registerRecordLoader("RecordLoaderTSM", "SleePyPhases.recordloaders")
        tsmPath = self.getConfig("tsm-path", False)
        if tsmPath is not False:
            tsmPath = Path(tsmPath)
            self.project.setConfig("loader.tsm.filePath", tsmPath.as_posix())
            self.project.setConfig("loader.tsm.dataset.downloader.basePath", tsmPath.as_posix())

        tsmPath = self.getConfig("tsm-domino-path", False)
        if tsmPath is not False:
            tsmPath = Path(tsmPath)
            self.project.setConfig("loader.tsm-domino.filePath", tsmPath.as_posix())
            self.project.setConfig("loader.tsm-domino.dataset.downloader.basePath", tsmPath.as_posix())

        preprocessingConfig = self.getConfig("preprocessing")
        if "combineChannels" in preprocessingConfig:
            RecordLoader.get().addCombinedChannels(preprocessingConfig["combineChannels"])


        # default fold is 0 if none is set
        self.setConfig("fold", self.getConfig("fold", 0))
        # self.setConfig("datasetSplits", list(self.getConfig("dataversion.split", {}).keys()))
        self.setConfig("hasFolds", self.getConfig("dataversion.folds", 0) > 0)

        s = Scorer(5)
        s.ignoreClasses = [-1]
        
        def reduceKappa(y_true, y_pred):
            # sleep stages are calculated for 50Hz, reduce to 30s for scoring
            y_pred = y_pred.reshape(-1, 1500, 5)
            y_pred = y_pred.mean(axis=1)
            y_true = y_true[::1500]
            return s.getMetricScorer("kappa")(y_true, y_pred)

        def combineReduceKappa(y_true, y_pred):
            s.reset()
            y_true, y_pred = s.maskedIgnoredValues(y_true, y_pred)
            kappa = reduceKappa(y_true, y_pred)
            print("######SleepStages######")
            print(s.results["confusion"])
            return kappa
        Scorer.registerMetric("reduceAndKappa", reduceKappa, combineReduceKappa)
        
        MedicalDB.config = self.getConfig("medicalDB", {})

            
    def getDataVersionManager(self):
        groupedRecords = self.project.getData("allDBRecordIds", list)
        seed = self.getConfig("dataversion.seed", None)
        splits = self.getConfig("dataversion.split", {})
        splits = {} if splits is None else splits


        # check if there are removed records, this can only happen after the extraction
        # dataversionamnager has diffent records before and after extraction!
        with self.project:
            if self.project.dataExistIn("removedRecordIds", list):
                removedRecordIds = self.getData("removedRecordIds", list, generate=False)
                self.log("Removing incomplete records from dataversionmanager")
                groupedRecords = {k: [r for r in v if r not in removedRecordIds] for k, v in groupedRecords.items()}
                # remove empty groups
                groupedRecords = {k: v for k, v in groupedRecords.items() if len(v) > 0}
            else:
                self.logError("no removed records specified! The splits might differ")
        
        # create non manual splits, if not specified
        valSplit = self.getConfig("validationSplit", 0.2)
        testSplit = self.getConfig("testSplit", 0.2)

        dm = DataversionManager(groupedRecords, seed=seed)

        # add manual addes splits
        for splitName in ["training", "validation", "test"]:
            if splitName in splits:
                dm.addSplitBySlices(splitName, splits[splitName])

        # add training folding splits
        if "trainval" in splits:
            foldcount = self.getConfig("dataversion.folds", 0)
            if foldcount > 0:
                dm.addSplitsByFold("training", "validation", splits["trainval"], self.getConfig("dataversion.folds", 0), self.getConfig("fold", 0))
            else:
                dm.addSplitByRemaining("validation", valSplit) # , remainingSplit=trainingSlice
                dm.addSplitByRemaining("training", 1)

            if "training" in splits or "validation" in splits:
                raise Exception("trainval and training/validation split are mutually exclusive. Please remove one of them from the config. This can be caused by loading multiple configs with different splits.")

        if "trainvaltest" in splits:
            foldCount = self.getConfig("dataversion.folds")
            currentFold = self.getConfig("fold", 0)
            testSlice, trainingSlice = dm.getSplitsByPercentage(
                dm.getRemainingSplit(), subLength=1 / foldCount, subPosition=currentFold
            )

            dm.addSplitBySlices("test", testSlice)

            dm.addSplitByRemaining("validation", valSplit) # , remainingSplit=trainingSlice
            dm.addSplitByRemaining("training", 1)

            

        if "test" not in dm.splits:
            dm.addSplitByRemaining("test", testSplit)

        if "validation" not in dm.splits:
            dm.addSplitByRemaining("validation", valSplit)

        if "training" not in dm.splits:
            dm.addSplitByRemaining("training", 1)

        dm.validatDatasetVersion(raiseException=self.getConfig("validatDataset", True))

        return dm
    
    def generateData(self, name):
        if name == "dataversionmanager":
            dm = self.getDataVersionManager()
            self.registerData("dataversionmanager", dm, save=False)



import numpy as np
from pyPhases import Phase


class DataAnalysis(Phase):
    """analyse data based on the preprocessed data, and data splits"""

    useMultiThreading = True

    def analyseLabelChannel(self, labelchannel):
        classCount = len(self.stats["classes"])
        segmentClasses = np.zeros(classCount, dtype=np.int32)

        classCount = np.unique(labelchannel, return_counts=True)
        examples = {}
        for i, count in zip(*classCount):
            segmentClasses[int(i)] = count

            # i = 2
            # where the class occurs
            occurrences = np.where(labelchannel == i)[0]
            # remove ascending numbers (only starting points left)
            diff = np.diff(occurrences)
            mask = np.hstack(([True], diff > 1))
            occurrences = occurrences[mask]

            # random starting point
            start = np.random.choice(occurrences)
            # find endpoint
            occurrencesOthers = np.where(labelchannel != i)[0]
            end = occurrencesOthers[occurrencesOthers > start]
            end = end[0] if len(end) > 0 else len(labelchannel)

            start = max(start, 0)
            end = min(end, len(labelchannel))
            examples[int(i)] = (start, end)

        return {
            "segmentClasses": segmentClasses,
            "classExamples": examples,
        }

    def analyseDataset(self, datasetName):
        classCount = len(self.stats["classes"])
        self.stats["classCounts"] = np.zeros(classCount)

        # analyse labels
        data = self.getData(
            "data-features",
            np.memmap,
            options={
                "dtype": self.getConfig("preprocessing.dtype"),
            },
        )

        recordsStats = []
        for record in data:
            # only get stats for last label channel (if ther multiple)
            recordStat = self.analyseLabelChannel(record[:, -1])
            self.stats["classCounts"] += recordStat["segmentClasses"]

            recordsStats.append(recordStat)

        self.stats["records"] = recordsStats

    def generateData(self, name):
        datasetName = self.getConfig("datasetSplit")
        labelNames = self.getConfig("classification.labelNames")
        self.stats = {"labels": labelNames}

        self.analyseDataset(datasetName)
        self.registerData("datasetstats", self.stats)
        self.logSuccess("Stats: %s" % self.stats)

    def main(self):
        analyseDatasets = self.getConfig("datasetSplits")

        for datasetName in analyseDatasets:
            self.setConfig("datasetSplit", datasetName)
            self.generateData("datasetstats")



import random
import numpy as np
import pandas as pd
from SleePyPhases.MultiScorer import MultiScorer
from pyPhases import Phase
from pyPhasesML import Model, ModelManager, Scorer
from tqdm import tqdm

from SleePyPhases.DataManipulation import DataManipulation
from SleePyPhases.EventScorer import EventScorer
from SleePyPhases.SleepMetaData import SleepMetaData


class Eval(Phase):
    allowTraining: bool = False
    exampleCount: int = 3

    def getModel(self) -> Model:
        modelState = self.project.getData("modelState", Model, generate=self.allowTraining)
        model = ModelManager.getModel(True)
        model.build()
        model.loadState(modelState)
        return model

    def getTestMetadata(self):
        if not bool(self.getConfig("evalOn", {})):
            return self.getData("metadata", list)

        with self.project:
            evalConfig = self.getConfig("evalOn")

            # try not to update dataversion, rather replace it completly
            if "dataversion" in evalConfig:
                self.project.config["dataversion"] = {}

            self.project.config.update(evalConfig)
            self.project.trigger("configChanged", None)
            dm = self.getData("metadata", list)

        return dm

    def getExternalTestset(self):
        with self.project:
            evalConfig = self.getConfig("evalOn")

            # try not to update dataversion, rather replace it completly
            if "dataversion" in evalConfig:
                self.project.config["dataversion"] = {}

            self.project.config.update(evalConfig)
            self.project.trigger("configChanged", None)
            split = self.getConfig("datasetSplit")
            dataset = self.project.getData(f"dataset-{split}", list)
            dm = self.getData("dataversionmanager")
            recordsMap = dm.getRecordsForSplit(split)

            # self.setConfig("datasetSplits", list(self.getConfig("dataversion.split", {}).keys()))
        return dataset, dm, recordsMap

    def getTestDataAndRecordMap(self):
        # load or generate test data and model state
        if bool(self.getConfig("evalOn", {})):
            testData, dm, recordsMap = self.getExternalTestset()
        else:
            testData = self.project.getData("dataset-test", list)
            dm = self.getData("dataversionmanager")
            recordsMap = dm.getRecordsForSplit("test")

        return testData, recordsMap

    def getTestData(self):
        return self.getTestDataAndRecordMap()[0]

    def getTestRecordMap(self):
        return self.getTestDataAndRecordMap()[1]

    def segmentEvaluation(self, name):
        testData = self.getTestData()
        self.setConfig("datasetSplits", list(self.getConfig("dataversion.split", {}).keys()))
        model = self.getModel()

        # setup data and scorer
        
        classNums = self.getConfig("model.multiclassificationSize", [2])
        labelNames = self.getConfig("classification.labelNames")
        ignoreIndex = self.getConfig("classification.ignoreIndex", -1)
        metrics = self.getConfig("trainingParameter.validationMetrics")


        scorer = MultiScorer(classNums, metrics, scorerNames=labelNames, ignoreClasses=[ignoreIndex], scorerClass=EventScorer)
        for s in scorer.scorer.values():
            s.trace = True
        threshold = self.getConfig("fixedThreshold", False)
        threshold = threshold or self.getData("threshold", list)
        scorer.setThresholds(threshold)

        manipulationAfterPredict = self.getConfig("manipulationAfterPredict", False)
        da = DataManipulation.getInstance(manipulationAfterPredict, "test", self.project.config, threshold=threshold)
        batchSize = self.getConfig("eval.batchSize")

        for data in tqdm(testData):
            x, truth = data
            # make sure to have a batchsize if batchSize == 1
            x = x.reshape(batchSize, x.shape[0], x.shape[1])
            truth = truth.reshape(batchSize, truth.shape[0], truth.shape[1])

            prediction = model.predict(x, get_likelihood=True, returnNumpy=False)
            prediction, truth = da((prediction, truth))

            prediction = prediction.reshape(-1, prediction.shape[2])
            truth = truth.reshape(-1, truth.shape[2])

            scorer.score(truth, prediction, trace=True)

        result = scorer.scoreAllRecords()

        return scorer.recordResult, result

    def getExample(self, eventstream):
        diff = np.diff(eventstream.astype(int), prepend=0, append=0)
        starts = np.where(diff == 1)[0]
        ends = np.where(diff == -1)[0] - 1
        index = random.randint(0, len(starts) - 1)
        return starts[index], ends[index]+1


    def getExamples(self, labelIndex, truth, predictions):
        # labelNames =
        labelName = self.getConfig("classification.labelNames")[labelIndex]
        classNames = self.getConfig("classification.classNames")[labelIndex]

        examples = {}
        truth, predictions = truth.reshape(-1), predictions.reshape(-1)

        for i, name in enumerate(classNames):
            for j, name in enumerate(classNames):
                exampleName = f"{labelName}-{i}-{j}"
                examples[exampleName] = []
                eventStream = (truth == i) & (predictions == j)
                if any(eventStream):
                    for e in range(self.exampleCount):
                        # example = self.getExample(occurrences, occurrencesOthers, len(truth))
                        example = self.getExample(eventStream)
                        examples[exampleName].append(example)

        return examples

    def eventEvaluation(self):
        # recordResults, segmentResults = self.getData("evalResults", list)
        testData, recordsMap = self.getTestDataAndRecordMap()

        with self.project:
            self.project.config.update(self.getConfig("eventEval"))

            evalMetrics = self.getConfig("eval.metrics")
            classNums = self.getConfig("model.multiclassificationSize", [2])
            labelNames = self.getConfig("classification.labelNames")
            predictionNames = self.getConfig("classification.predictionSignals")
            predictionFrequencies = self.getConfig("classification.predictionFrequencies")
            ignoreIndex = self.getConfig("classification.ignoreIndex", -1)

            threshold = self.getConfig("fixedThreshold", False)
            thresholds = threshold or self.getData("threshold", list)
            mScorer = MultiScorer(classNums, evalMetrics, scorerNames=labelNames, ignoreClasses=[ignoreIndex], scorerClass=EventScorer)
            for threshold, s in zip(thresholds, mScorer.scorer.values()):
                s.majorityVote = self.getConfig("eventEval.tpStrat", "overlap") == "majority"
                s.noTN = self.getConfig("eventEval.tnStrat", "eventcount") == "noTN"
                s.threshold = threshold

            manipulationAfterPredict = self.getConfig("manipulationAfterPredict", False)
            da = DataManipulation.getInstance(manipulationAfterPredict, "test", self.project.config, threshold=thresholds)

            metaData = pd.DataFrame(self.getTestMetadata())
            metaData.replace("M", None, inplace=True)
            metaData["indexArousal"] = metaData["indexArousal"].astype(float)
            numClasses = self.getConfig("numClasses")

            model = self.getModel()
            
            resultRows = []
            result = {}
            # for recordIndex in tqdm(range(recordCount)):
            for recordIndex, data in tqdm(enumerate(testData), total=len(testData)):
                x, truth = data

                recordId = recordsMap[recordIndex]
                resultRow = {"recordId": recordId}

                sleepDataTruth = SleepMetaData()
                sleepDataPredicted = SleepMetaData()

                if len(x.shape) == 2:
                    x = x.reshape([1] + list(x.shape))
                    truth = truth.reshape([1] + list(truth.shape))

                prediction = model.predict(x, get_likelihood=True, returnNumpy=False)
                prediction, truth = da((prediction, truth))

                # for labelIndex, labelName in enumerate(recordResults.keys()):
                for labelIndex, labelName in enumerate(labelNames):
                    predictionName = predictionNames[labelIndex]
                    predictionFrequency = predictionFrequencies[labelIndex]

                    # recordResult = recordResults[labelName][recordIndex]
                    numClass = numClasses[labelIndex]

                    scorer = mScorer.scorer[labelName]
                    labelPrediction = prediction[labelIndex] if isinstance(prediction, list) else prediction[:, :, labelIndex]
                    labelTruth = truth[labelIndex] if isinstance(truth, list) else truth[:, :, labelIndex]

                    predictionSignal = labelPrediction.reshape(-1)
                    labelTruth = labelTruth.reshape(-1)

                    # score the prediction
                    r = scorer.score(labelTruth.reshape(-1), labelPrediction.reshape(-1, numClass))
                    resultRow.update({f"{m}_{labelName}": v for m,v in r.items()})

                    # add examples
                    resultRow[f"examples_{labelName}"] = self.getExamples(labelIndex, labelTruth, labelPrediction)


                    recordData = metaData.query(f'recordId == "{recordId}"')
                    recordData = recordData.iloc[0] if len(recordData) == 1 else {}
                    sleepDataTruth.addSignal(predictionName, labelTruth, frequency=predictionFrequency)
                    sleepDataPredicted.addSignal(predictionName, scorer.flattenPrediction(predictionSignal), frequency=predictionFrequency)

                for metricName in sleepDataPredicted.metaDataMap.keys():
                    resultRow[f"{metricName}-software"] = (
                        recordData[metricName].item()
                        if metricName in recordData and bool(recordData[metricName].tolist())
                        else 0
                    )
                    resultRow[f"{metricName}-truth"] = sleepDataTruth.getMetaData(metricName)
                    resultRow[f"{metricName}-prediction"] = sleepDataPredicted.getMetaData(metricName)

                resultRows.append(resultRow)
            
            for name, scorer in mScorer.scorer.items():
                allLabelResults = scorer.scoreAllRecords()
                result.update({f"{m}_{name}": v for m,v in allLabelResults.items()})
        
        return resultRows, result

    def generateData(self, name):
        if name == "evalResults":
            segmentResults = self.segmentEvaluation(name)
            self.registerData("evalResults", segmentResults)
        elif name == "eventResults":
            eventResults = self.eventEvaluation()
            self.registerData("eventResults", eventResults)

    def main(self):
        self.getData("eventResults", list)
